#!/usr/bin/env python3
#
# Copyright (c) 2022 Jan Behrens
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

import random
import math

# helper function, which re-distributes counts such that they do not exceed
# nearest integers of the actual target values
#
# targets: dict mapping category to ideal target count
# lows: dict with lowered target counts (due to rounding of parent category)
# highs: dict with raised target counts (due to rounding of parent category)
# high_prob: probability of using raised target counts
def catpick_distribute(targets, lows, highs, high_prob):
    # calculate boundaries based on targets
    low_bounds = {}
    high_bounds = {}
    for cat, target in targets.items():
        low_bounds[cat] = math.floor(target)
        high_bounds[cat] = math.ceil(target)
    # categories which can still be adjusted
    open_cats = set(targets)
    while True:
        # errors (positive numbers) introduced in lows or highs
        low_error = 0
        high_error = 0
        # categories to be removed from open_cats after iteration
        cats_to_close = []
        # limit counts in categories which may still be adjusted
        # and remember errors caused
        for cat in open_cats:
            error = low_bounds[cat] - lows[cat]
            if error > 0:
                lows[cat] = low_bounds[cat]
                low_error += error
                other_error = error * ((1-high_prob) / high_prob)
                highs[cat] -= other_error
                high_error += other_error
                cats_to_close.append(cat)
                continue
            error = highs[cat] - high_bounds[cat]
            if error > 0:
                highs[cat] = high_bounds[cat]
                high_error += error
                other_error = error * (high_prob / (1-high_prob))
                lows[cat] += other_error
                low_error += other_error
                cats_to_close.append(cat)
        # stop when there is no more error to be distributed
        if low_error == 0 and high_error == 0: break
        # mark categories which were force-limited as done
        for cat in cats_to_close:
            open_cats.remove(cat)
        # stop when all categories are done
        if not open_cats: break
        # distribute error
        low_total = 0
        high_total = 0
        for cat in open_cats:
            low_total += lows[cat]
            high_total += highs[cat]
        for cat in open_cats:
            if low_total: lows[cat] -= low_error * lows[cat] / low_total
            if high_total: highs[cat] += high_error * highs[cat] / high_total

# recursive helper function discretizing hierarchical counts
#
# parent: parent category as tuple
# parent_count: desired count for parent category (may be non-integer)
# round_up: boolean determining if parent_count is rounded up (or down)
# in_counts: dict mapping leaf-categories to counts
# out_counts: output dict (categories mapped to counts, may be sparse)
def catpick_discretize(parent, parent_count, round_up, in_counts, out_counts):

    # perform rounding and calculate rounding probability
    parent_count_floor = math.floor(parent_count)
    parent_count_ceil = math.ceil(parent_count)
    parent_count_frac = parent_count % 1
    if round_up:
        parent_count_rounded = parent_count_ceil
    else:
        parent_count_rounded = parent_count_floor

    # avoid division by zero
    if parent_count_rounded == 0: return

    # output rounded count for current category
    out_counts[parent] = parent_count_rounded

    # determine hierarchy level
    level = len(parent)

    # determine childen and their counts as well as the total count
    children = {}
    total = 0
    for cat, count in in_counts.items():
        if len(cat) > level and cat[0:level] == parent:
            child = cat[0:level+1]
            children[child] = children.get(child, 0) + count
            total += count

    # check if done
    if total == 0: return

    # counts to pick per category
    # (real target, and counts in case of rounding count down or up)
    targets, lows, highs = {}, {}, {}
    for cat, count in children.items():
        target = parent_count * count / total
        targets[cat] = target
        lows[cat] = target * parent_count_floor / parent_count
        highs[cat] = target * parent_count_ceil / parent_count

    # limit counts to nearest integer
    catpick_distribute(targets, lows, highs, parent_count_frac)

    # target counts for categories based on demanded rounding
    if round_up: targets = highs
    else: targets = lows

    # round target counts to integers
    rounded_counts = {}
    done_counter = 0
    remaining_targets = dict(targets)
    while done_counter < parent_count_rounded:
        cat, target = remaining_targets.popitem()
        if not remaining_targets:
            rounded_counts[cat] = parent_count_rounded - done_counter
            break
        round_up_error = 1 - (target % 1)
        round_down_error = target % 1
        remaining_total = 0
        for remaining_cat, remaining_target in remaining_targets.items():
            remaining_total += remaining_target
        lows, highs = {}, {}
        for remaining_cat, remaining_target in remaining_targets.items():
            if remaining_total > 0:
                lows[remaining_cat] = (
                    remaining_target -
                    round_up_error * remaining_target / remaining_total
                )
                highs[remaining_cat] = (
                    remaining_target +
                    round_down_error * remaining_target / remaining_total
                )
            else:
                lows[remaining_cat] = math.floor(remaining_target + 0.5)
                highs[remaining_cat] = math.floor(remaining_target + 0.5)
        catpick_distribute(remaining_targets, lows, highs, round_up_error)
        if round_up_error > random.random():
            remaining_targets = highs
            rounded_count = math.floor(target)
        else:
            remaining_targets = lows
            rounded_count = math.ceil(target)
        rounded_counts[cat] = rounded_count
        done_counter += rounded_count
    for cat in remaining_targets:
        rounded_counts[cat] = 0

    # process children
    for cat in children:
        catpick_discretize(
            cat,
            targets[cat],
            rounded_counts[cat] > targets[cat],
            in_counts,
            out_counts
        )

# recursive helper function determining counts for one element in batch
#
# parent: parent category as tuple
# parent_count: desired integer count for parent category
# single_counts: dict mapping leaf-categories to ideal counts
# batch_counts: sparse dict mapping categories to remaining counts for batch
# batch_remaining: remaining elements for batch
# out_counts: output dict (categories mapped to counts)
def catpick_batch(
    parent,
    parent_count,
    single_counts,
    batch_counts,
    batch_remaining,
    out_counts
):

    # move count for parent category from batch_counts to out_counts
    if parent in batch_counts: batch_counts[parent] -= parent_count
    out_counts[parent] = parent_count

    # determine ideal counts for children
    targets = {}
    level = len(parent)
    for cat in single_counts:
        if len(cat) > level and cat[0:level] == parent:
            child = cat[0:level+1]
            targets[child] = (
                targets.get(child, 0) + single_counts[cat]
            )

    # stop of there are no children
    if not targets: return

    # determine ideal counts for remaining batch
    batch_targets = {}
    for cat in targets:
        batch_targets[cat] = batch_counts.get(cat, 0) / batch_remaining

    # round counts for children based on ideal counts for remaining batch
    rounded_targets = {}
    for cat, target in targets.items():
        target_floor = math.floor(target)
        target_ceil = math.ceil(target)
        target_middle = (target_floor + target_ceil) / 2
        if target_middle < batch_targets[cat]:
            rounded_targets[cat] = math.ceil(target)
        else:
            rounded_targets[cat] = math.floor(target)

    # check if sum of children count is correct and adjust rounding if it's not
    total = 0
    for cat, target in rounded_targets.items():
        total += target
    while total < parent_count:
        best_error = None
        candidates = {}
        for cat, target in rounded_targets.items():
            # check if category count can be rounded up
            if targets[cat] > target:
                error = (target + 1) - batch_targets[cat]
                if best_error == None or best_error > error:
                    best_error = error
                candidates[cat] = error
        best_candidates = []
        for candidate, error in candidates.items():
            if error == best_error:
                best_candidates.append(candidate)
        rounded_targets[random.choice(best_candidates)] += 1
        total += 1
    while total > parent_count:
        best_error = None
        candidates = {}
        for cat, target in rounded_targets.items():
            # check if category count can be rounded down
            if targets[cat] < target:
                error = batch_targets[cat] - (target - 1)
                if best_error == None or best_error > error:
                    best_error = error
                candidates[cat] = error
        best_candidates = []
        for candidate, error in candidates.items():
            if error == best_error:
                best_candidates.append(candidate)
        rounded_targets[random.choice(best_candidates)] -= 1
        total -= 1

    # process children
    for cat, target in rounded_targets.items():
        catpick_batch(
            cat,
            target,
            single_counts,
            batch_counts,
            batch_remaining,
            out_counts
        )

# repeatedly pick a certain number of items from a pool while considering item
# categorization
#
# category_mapper: function mapping item to category (as sequence)
# count: number of items per sample
# batch_count: number of samples
def catpick(category_mapper, pool, count, batch_count):
    in_counts = {}
    subpools = {}
    for item in pool:
        cat = tuple(category_mapper(item))
        in_counts[cat] = in_counts.get(cat, 0) + 1
        if cat in subpools:
            subpools[cat].append(item)
        else:
            subpools[cat] = [item]
    single_counts = {}
    for cat, subpool in subpools.items():
        single_counts[cat] = len(subpool) * count / len(pool)
    total_count = count * batch_count
    batch_counts = {}
    catpick_discretize(
        (),
        total_count,
        total_count % 1 > random.random(),
        in_counts,
        batch_counts,
    )
    results = []
    for i in range(batch_count):
        out_counts = {}
        catpick_batch(
            (),
            math.floor(count + random.random()),
            single_counts,
            batch_counts,
            batch_count - i,
            out_counts
        )
        result = []
        for cat, subpool in subpools.items():
            sublen = len(subpool)
            remaining = out_counts.get(cat, 0)
            while remaining >= sublen:
                result += subpool
                remaining -= sublen
            result += random.sample(subpool, remaining)
        random.shuffle(result)
        results.append(result)
    random.shuffle(results)
    return results

# test case to show that program works correctly
if __name__ == "__main__":

    def rounding_variance(target):
        frac = target % 1
        compl = 1 - frac
        return frac*frac*compl + compl*compl*frac

    def assert_eq(left, right):
        if not abs(left - right) < 1e-12:
            raise AssertionError("testing assertion failed")
    def assert_range(value, low, high):
        if not (value >= low and value <= high):
            raise AssertionError("testing assertion failed")
    def assert_average(rounds, target, result):
        average = rounds * target
        stddev = math.sqrt(rounds * rounding_variance(target))
        allowed = 6 * stddev
        assert_range(result, average - allowed, average + allowed)

    def category_mapper(item):
        result = []
        for char in item:
            if char == "-": break
            result.append(char)
        return result

    for i in range(1, 1+10):
        batch = i % 2 == 0
        if not batch:
            print("Running test #" + str(i) + " of 10 (independent)...")
        else:
            print("Running test #" + str(i) + " of 10 (batch operation)...")
        target_a1 = random.randint(1, 30)
        target_a2 = random.randint(1, 30)
        target_a3 = random.randint(1, 30)
        target_b1 = random.randint(1, 12)
        target_b2 = random.randint(1, 12)
        target_b3 = random.randint(1, 12)
        target_c1 = random.randint(1, 6)
        target_c2 = random.randint(1, 6)
        target_c3 = random.randint(1, 6)
        target_a = target_a1 + target_a2 + target_a3
        target_b = target_b1 + target_b2 + target_b3
        target_c = target_c1 + target_c2 + target_c3
        count = target_a + target_b + target_c
        pool = []
        for q in range(len(pool)+1, len(pool)+1+target_a1):
            pool.append("A1-" + str(q))
        for q in range(len(pool)+1, len(pool)+1+target_a2):
            pool.append("A2-" + str(q))
        for q in range(len(pool)+1, len(pool)+1+target_a3):
            pool.append("A3-" + str(q))
        for q in range(len(pool)+1, len(pool)+1+target_b1):
            pool.append("B1-" + str(q))
        for q in range(len(pool)+1, len(pool)+1+target_b2):
            pool.append("B2-" + str(q))
        for q in range(len(pool)+1, len(pool)+1+target_b3):
            pool.append("B3-" + str(q))
        for q in range(len(pool)+1, len(pool)+1+target_c1):
            pool.append("C1-" + str(q))
        for q in range(len(pool)+1, len(pool)+1+target_c2):
            pool.append("C2-" + str(q))
        for q in range(len(pool)+1, len(pool)+1+target_c3):
            pool.append("C3-" + str(q))
        if i % 2 == 1:
            random.shuffle(pool)
        sample_count = random.randint(1, random.randint(1, count))
        target_a1 *= sample_count / count
        target_a2 *= sample_count / count
        target_a3 *= sample_count / count
        target_b1 *= sample_count / count
        target_b2 *= sample_count / count
        target_b3 *= sample_count / count
        target_c1 *= sample_count / count
        target_c2 *= sample_count / count
        target_c3 *= sample_count / count
        target_a *= sample_count / count
        target_b *= sample_count / count
        target_c *= sample_count / count
        target_a1_floor = math.floor(target_a1)
        target_a1_ceil = math.ceil(target_a1)
        target_a2_floor = math.floor(target_a2)
        target_a2_ceil = math.ceil(target_a2)
        target_a3_floor = math.floor(target_a3)
        target_a3_ceil = math.ceil(target_a3)
        target_b1_floor = math.floor(target_b1)
        target_b1_ceil = math.ceil(target_b1)
        target_b2_floor = math.floor(target_b2)
        target_b2_ceil = math.ceil(target_b2)
        target_b3_floor = math.floor(target_b3)
        target_b3_ceil = math.ceil(target_b3)
        target_c1_floor = math.floor(target_c1)
        target_c1_ceil = math.ceil(target_c1)
        target_c2_floor = math.floor(target_c2)
        target_c2_ceil = math.ceil(target_c2)
        target_c3_floor = math.floor(target_c3)
        target_c3_ceil = math.ceil(target_c3)
        target_a_floor = math.floor(target_a)
        target_a_ceil = math.ceil(target_a)
        target_b_floor = math.floor(target_b)
        target_b_ceil = math.ceil(target_b)
        target_c_floor = math.floor(target_c)
        target_c_ceil = math.ceil(target_c)
        rounds = 30000
        sum_a1, sum_a2, sum_a3 = 0, 0, 0
        sum_b1, sum_b2, sum_b3 = 0, 0, 0
        sum_c1, sum_c2, sum_c3 = 0, 0, 0
        sum_a, sum_b, sum_c = 0, 0, 0
        batch_samples = []
        for j in range(rounds):
            if batch:
                if not batch_samples:
                    batch_samples = catpick(
                        category_mapper,
                        pool,
                        sample_count,
                        20
                    )
                samples = batch_samples.pop()
            else:
                samples = catpick(category_mapper, pool, sample_count, 1)[0]
            count_a1 = len(list(filter(lambda x: x[0:2] == "A1", samples)))
            count_a2 = len(list(filter(lambda x: x[0:2] == "A2", samples)))
            count_a3 = len(list(filter(lambda x: x[0:2] == "A3", samples)))
            count_b1 = len(list(filter(lambda x: x[0:2] == "B1", samples)))
            count_b2 = len(list(filter(lambda x: x[0:2] == "B2", samples)))
            count_b3 = len(list(filter(lambda x: x[0:2] == "B3", samples)))
            count_c1 = len(list(filter(lambda x: x[0:2] == "C1", samples)))
            count_c2 = len(list(filter(lambda x: x[0:2] == "C2", samples)))
            count_c3 = len(list(filter(lambda x: x[0:2] == "C3", samples)))
            count_a = count_a1 + count_a2 + count_a3
            count_b = count_b1 + count_b2 + count_b3
            count_c = count_c1 + count_c2 + count_c3
            assert_eq(count_a + count_b + count_c, sample_count)
            assert_range(count_a1, target_a1_floor, target_a1_ceil)
            assert_range(count_a2, target_a2_floor, target_a2_ceil)
            assert_range(count_a3, target_a3_floor, target_a3_ceil)
            assert_range(count_b1, target_b1_floor, target_b1_ceil)
            assert_range(count_b2, target_b2_floor, target_b2_ceil)
            assert_range(count_b3, target_b3_floor, target_b3_ceil)
            assert_range(count_c1, target_c1_floor, target_c1_ceil)
            assert_range(count_c2, target_c2_floor, target_c2_ceil)
            assert_range(count_c3, target_c3_floor, target_c3_ceil)
            assert_range(count_a, target_a_floor, target_a_ceil)
            assert_range(count_b, target_b_floor, target_b_ceil)
            assert_range(count_c, target_c_floor, target_c_ceil)
            sum_a1 += count_a1; sum_a2 += count_a2; sum_a3 += count_a3
            sum_b1 += count_b1; sum_b2 += count_b2; sum_b3 += count_b3
            sum_c1 += count_c1; sum_c2 += count_c2; sum_c3 += count_c3
        sum_a = sum_a1 + sum_a2 + sum_a3
        sum_b = sum_b1 + sum_b2 + sum_b3
        sum_c = sum_c1 + sum_c2 + sum_c3
        assert_average(rounds, target_a1, sum_a1)
        assert_average(rounds, target_a2, sum_a2)
        assert_average(rounds, target_a3, sum_a3)
        assert_average(rounds, target_b1, sum_b1)
        assert_average(rounds, target_b2, sum_b2)
        assert_average(rounds, target_b3, sum_b3)
        assert_average(rounds, target_c1, sum_c1)
        assert_average(rounds, target_c2, sum_c2)
        assert_average(rounds, target_c3, sum_c3)
        assert_average(rounds, target_a, sum_a)
        assert_average(rounds, target_b, sum_b)
        assert_average(rounds, target_c, sum_c)
    print("All tests passed.")
